@@ -23,12 +23,12 @@
 
 #include <config.h>
 
-#include <deque>
 #include <memory>
 #include <set>
 #include <stack>
 #include <string.h>
 #include <tuple>
+#include <unordered_map>
 #include <vector>
 
 #include "object.h"
@@ -52,7 +52,6 @@
 #include "gjs/mem.h"
 
 #include <util/log.h>
-#include <util/hash-x32.h>
 #include <girepository.h>
 
 struct ObjectInstance {
@@ -61,29 +60,27 @@ struct ObjectInstance {
     GjsMaybeOwned<JSObject *> keep_alive;
     GType gtype;
 
-    /* a list of all signal connections, used when tracing */
-    std::set<GClosure *> signals;
+    /* a list of all GClosures installed on this object (from
+     * signals, trampolines and explicit GClosures), used when tracing */
+    std::set<GClosure *> closures;
 
     /* the GObjectClass wrapped by this JS Object (only used for
        prototypes) */
     GTypeClass *klass;
 
-    /* A list of all vfunc trampolines, used when tracing */
-    std::deque<GjsCallbackTrampoline *> vfuncs;
-
     unsigned js_object_finalized : 1;
 };
 
 static std::stack<JS::PersistentRootedObject> object_init_list;
-static GHashTable *class_init_properties;
+
+using ParamRef = std::unique_ptr<GParamSpec, decltype(&g_param_spec_unref)>;
+using ParamRefArray = std::vector<ParamRef>;
+static std::unordered_map<GType, ParamRefArray> class_init_properties;
 
 static bool weak_pointer_callback = false;
-static std::set<ObjectInstance *> weak_pointer_list;
+static std::set<ObjectInstance *> wrapped_gobject_list;
 
 extern struct JSClass gjs_object_instance_class;
-
-static std::set<ObjectInstance *> dissociate_list;
-
 GJS_DEFINE_PRIV_FROM_JS(ObjectInstance, gjs_object_instance_class)
 
 static void            disassociate_js_gobject (GObject *gobj);
@@ -148,21 +145,6 @@ throw_priv_is_null_error(JSContext *context)
               " up to the parent _init properly?");
 }
 
-static void
-dissociate_list_add(ObjectInstance *priv)
-{
-    bool inserted;
-    std::tie(std::ignore, inserted) = dissociate_list.insert(priv);
-    g_assert(inserted);
-}
-
-static void
-dissociate_list_remove(ObjectInstance *priv)
-{
-    size_t erased = dissociate_list.erase(priv);
-    g_assert(erased > 0);
-}
-
 static ObjectInstance *
 get_object_qdata(GObject *gobj)
 {
@@ -938,7 +920,7 @@ static void
 wrapped_gobj_dispose_notify(gpointer      data,
                             GObject      *where_the_object_was)
 {
-    weak_pointer_list.erase(static_cast<ObjectInstance *>(data));
+    wrapped_gobject_list.erase(static_cast<ObjectInstance *>(data));
 #if DEBUG_DISPOSE
     gjs_debug(GJS_DEBUG_GOBJECT, "Wrapped GObject %p disposed", where_the_object_was);
 #endif
@@ -957,8 +939,7 @@ gobj_no_longer_kept_alive_func(JS::HandleObject obj,
                         obj.get());
 
     priv->keep_alive.reset();
-    dissociate_list_remove(priv);
-    weak_pointer_list.erase(priv);
+    wrapped_gobject_list.erase(priv);
 }
 
 static void
@@ -976,7 +957,6 @@ handle_toggle_down(GObject *gobj)
     if (priv->keep_alive.rooted()) {
         gjs_debug_lifecycle(GJS_DEBUG_GOBJECT, "Removing object from keep alive");
         priv->keep_alive.switch_to_unrooted();
-        dissociate_list_remove(priv);
     }
 }
 
@@ -1006,7 +986,6 @@ handle_toggle_up(GObject   *gobj)
         gjs_debug_lifecycle(GJS_DEBUG_GOBJECT, "Adding object to keep alive");
         auto cx = static_cast<JSContext *>(gjs_context_get_native_context(context));
         priv->keep_alive.switch_to_rooted(cx, gobj_no_longer_kept_alive_func, priv);
-        dissociate_list_add(priv);
     }
 }
 
@@ -1141,9 +1120,18 @@ gjs_object_prepare_shutdown(void)
      *   toggle ref removal -> gobj dispose -> toggle ref notify
      * by simply ignoring toggle ref notifications during this process.
      */
-    for (auto iter : dissociate_list)
-        release_native_object(iter);
-    dissociate_list.clear();
+    std::vector<ObjectInstance *> to_be_released;
+    for (auto iter = wrapped_gobject_list.begin(); iter != wrapped_gobject_list.end(); ) {
+        ObjectInstance *priv = *iter;
+        if (priv->keep_alive.rooted()) {
+            to_be_released.push_back(priv);
+            iter = wrapped_gobject_list.erase(iter);
+        } else {
+            iter++;
+        }
+    }
+    for (ObjectInstance *priv : to_be_released)
+        release_native_object(priv);
 }
 
 static ObjectInstance *
@@ -1186,7 +1174,7 @@ update_heap_wrapper_weak_pointers(JSContext     *cx,
 {
     std::vector<GObject *> to_be_disassociated;
 
-    for (auto iter = weak_pointer_list.begin(); iter != weak_pointer_list.end(); ) {
+    for (auto iter = wrapped_gobject_list.begin(); iter != wrapped_gobject_list.end(); ) {
         ObjectInstance *priv = *iter;
         if (priv->keep_alive.rooted() || priv->keep_alive == nullptr ||
             !priv->keep_alive.update_after_gc()) {
@@ -1198,7 +1186,7 @@ update_heap_wrapper_weak_pointers(JSContext     *cx,
              * may also cause it to be erased.)
              */
             to_be_disassociated.push_back(priv->gobj);
-            iter = weak_pointer_list.erase(iter);
+            iter = wrapped_gobject_list.erase(iter);
         }
     }
 
@@ -1232,7 +1220,7 @@ associate_js_gobject (JSContext       *context,
     set_object_qdata(gobj, priv);
 
     ensure_weak_pointer_callback(context);
-    weak_pointer_list.insert(priv);
+    wrapped_gobject_list.insert(priv);
 
     g_object_weak_ref(gobj, wrapped_gobj_dispose_notify, priv);
 
@@ -1248,22 +1236,21 @@ associate_js_gobject (JSContext       *context,
      * wrappee).
      */
     priv->keep_alive.root(context, object, gobj_no_longer_kept_alive_func, priv);
-    dissociate_list_add(priv);
     g_object_add_toggle_ref(gobj, wrapped_gobj_toggle_notify, NULL);
 }
 
 static void
-invalidate_all_signals(ObjectInstance *priv)
+invalidate_all_closures(ObjectInstance *priv)
 {
     /* Can't loop directly through the items, since invalidating an item's
      * closure might have the effect of removing the item from the set in the
      * invalidate notifier */
-    while (!priv->signals.empty()) {
+    while (!priv->closures.empty()) {
         /* This will also free cd, through the closure invalidation mechanism */
-        GClosure *closure = *priv->signals.begin();
+        GClosure *closure = *priv->closures.begin();
         g_closure_invalidate(closure);
         /* Erase element if not already erased */
-        priv->signals.erase(closure);
+        priv->closures.erase(closure);
     }
 }
 
@@ -1292,7 +1279,7 @@ disassociate_js_gobject(GObject *gobj)
                    gobj, G_OBJECT_TYPE_NAME(gobj));
     }
 
-    invalidate_all_signals(priv);
+    invalidate_all_closures(priv);
     release_native_object(priv);
 
     /* Mark that a JS object once existed, but it doesn't any more */
@@ -1434,19 +1421,16 @@ object_instance_trace(JSTracer *tracer,
     if (priv == NULL)
         return;
 
-    for (GClosure *closure : priv->signals)
+    for (GClosure *closure : priv->closures)
         gjs_closure_trace(closure, tracer);
-
-    for (auto vfunc : priv->vfuncs)
-        vfunc->js_function.trace(tracer, "ObjectInstance::vfunc");
 }
 
 static void
-signal_connection_invalidated(void     *data,
-                              GClosure *closure)
+closure_invalidated(void     *data,
+                    GClosure *closure)
 {
     auto priv = static_cast<ObjectInstance *>(data);
-    priv->signals.erase(closure);
+    priv->closures.erase(closure);
 }
 
 static void
@@ -1470,7 +1454,7 @@ object_instance_finalize(JSFreeOp  *fop,
 
     /* This applies only to instances, not prototypes, but it's possible that
      * an instance's GObject is already freed at this point. */
-    invalidate_all_signals(priv);
+    invalidate_all_closures(priv);
 
     /* Object is instance, not prototype, AND GObject is not already freed */
     if (priv->gobj) {
@@ -1495,12 +1479,6 @@ object_instance_finalize(JSFreeOp  *fop,
         release_native_object(priv);
     }
 
-    /* We have to leak the trampolines, since the GType's vtable still refers
-     * to them */
-    for (auto iter : priv->vfuncs)
-        iter->js_function.reset();
-    priv->vfuncs.clear();
-
     if (priv->keep_alive.rooted()) {
         /* This happens when the refcount on the object is still >1,
          * for example with global objects GDK never frees like GdkDisplay,
@@ -1513,8 +1491,8 @@ object_instance_finalize(JSFreeOp  *fop,
                             "Removing from keep alive");
 
         priv->keep_alive.reset();
-        dissociate_list_remove(priv);
     }
+    wrapped_gobject_list.erase(priv);
 
     if (priv->info) {
         g_base_info_unref( (GIBaseInfo*) priv->info);
@@ -1559,7 +1537,9 @@ gjs_lookup_object_constructor_from_info(JSContext    *context,
         /* In case we're looking for a private type, and we don't find it,
            we need to define it first.
         */
-        gjs_define_object_class(context, in_object, NULL, gtype, &constructor);
+        JS::RootedObject ignored(context);
+        gjs_define_object_class(context, in_object, NULL, gtype, &constructor,
+                                &ignored);
     } else {
         if (G_UNLIKELY (!value.isObject()))
             return NULL;
@@ -1609,6 +1589,16 @@ gjs_lookup_object_prototype(JSContext *context,
     return proto;
 }
 
+static void
+do_associate_closure(ObjectInstance *priv,
+                     GClosure       *closure)
+{
+    /* This is a weak reference, and will be cleared when the closure is
+     * invalidated */
+    priv->closures.insert(closure);
+    g_closure_add_invalidate_notifier(closure, priv, closure_invalidated);
+}
+
 static bool
 real_connect_func(JSContext *context,
                   unsigned   argc,
@@ -1658,11 +1648,7 @@ real_connect_func(JSContext *context,
     closure = gjs_closure_new_for_signal(context, &argv[1].toObject(), "signal callback", signal_id);
     if (closure == NULL)
         return false;
-
-    /* This is a weak reference, and will be cleared when the closure is invalidated */
-    priv->signals.insert(closure);
-    g_closure_add_invalidate_notifier(closure, priv,
-                                      signal_connection_invalidated);
+    do_associate_closure(priv, closure);
 
     id = g_signal_connect_closure_by_id(priv->gobj,
                                         signal_id,
@@ -1929,10 +1915,11 @@ gjs_define_object_class(JSContext              *context,
                         JS::HandleObject        in_object,
                         GIObjectInfo           *info,
                         GType                   gtype,
-                        JS::MutableHandleObject constructor)
+                        JS::MutableHandleObject constructor,
+                        JS::MutableHandleObject prototype)
 {
     const char *constructor_name;
-    JS::RootedObject prototype(context), parent_proto(context);
+    JS::RootedObject parent_proto(context);
 
     ObjectInstance *priv;
     const char *ns;
@@ -2001,7 +1988,7 @@ gjs_define_object_class(JSContext              *context,
                                 NULL,
                                 /* funcs of constructor, MyConstructor.myfunc() */
                                 NULL,
-                                &prototype,
+                                prototype,
                                 constructor)) {
         g_error("Can't init class %s", constructor_name);
     }
@@ -2137,13 +2124,13 @@ gjs_typecheck_object(JSContext       *context,
 
     if (!result && throw_error) {
         if (priv->info) {
-            gjs_throw_custom(context, "TypeError", NULL,
+            gjs_throw_custom(context, JSProto_TypeError, nullptr,
                              "Object is of type %s.%s - cannot convert to %s",
                              g_base_info_get_namespace((GIBaseInfo*) priv->info),
                              g_base_info_get_name((GIBaseInfo*) priv->info),
                              g_type_name(expected_type));
         } else {
-            gjs_throw_custom(context, "TypeError", NULL,
+            gjs_throw_custom(context, JSProto_TypeError, nullptr,
                              "Object is of type %s - cannot convert to %s",
                              g_type_name(priv->gtype),
                              g_type_name(expected_type));
@@ -2307,30 +2294,20 @@ gjs_hook_up_vfunc(JSContext *cx,
 
     find_vfunc_info(cx, gtype, vfunc, name, &implementor_vtable, &field_info);
     if (field_info != NULL) {
-        GITypeInfo *type_info;
-        GIBaseInfo *interface_info;
-        GICallbackInfo *callback_info;
         gint offset;
         gpointer method_ptr;
         GjsCallbackTrampoline *trampoline;
 
-        type_info = g_field_info_get_type(field_info);
-
-        interface_info = g_type_info_get_interface(type_info);
-
-        callback_info = (GICallbackInfo*)interface_info;
         offset = g_field_info_get_offset(field_info);
         method_ptr = G_STRUCT_MEMBER_P(implementor_vtable, offset);
 
         JS::RootedValue v_function(cx, JS::ObjectValue(*function));
-        trampoline = gjs_callback_trampoline_new(cx, v_function, callback_info,
-                                                 GI_SCOPE_TYPE_NOTIFIED, true);
+        trampoline = gjs_callback_trampoline_new(cx, v_function, vfunc,
+                                                 GI_SCOPE_TYPE_NOTIFIED,
+                                                 object, true);
 
         *((ffi_closure **)method_ptr) = trampoline->closure;
-        priv->vfuncs.push_back(trampoline);
 
-        g_base_info_unref(interface_info);
-        g_base_info_unref(type_info);
         g_base_info_unref(field_info);
     }
 
@@ -2526,49 +2503,45 @@ static void
 gjs_interface_init(GTypeInterface *g_iface,
                    gpointer        iface_data)
 {
-    GPtrArray *properties;
-    GType gtype;
-    guint i;
+    GType gtype = G_TYPE_FROM_INTERFACE(g_iface);
 
-    gtype = G_TYPE_FROM_INTERFACE (g_iface);
-
-    properties = (GPtrArray *) gjs_hash_table_for_gsize_lookup(class_init_properties, gtype);
-    if (properties == NULL)
+    auto found = class_init_properties.find(gtype);
+    if (found == class_init_properties.end())
         return;
 
-    for (i = 0; i < properties->len; i++) {
-        GParamSpec *pspec = (GParamSpec *) properties->pdata[i];
-        g_param_spec_set_qdata(pspec, gjs_is_custom_property_quark(), GINT_TO_POINTER(1));
-        g_object_interface_install_property(g_iface, pspec);
+    ParamRefArray& properties = found->second;
+    for (ParamRef& pspec : properties) {
+        g_param_spec_set_qdata(pspec.get(), gjs_is_custom_property_quark(),
+                               GINT_TO_POINTER(1));
+        g_object_interface_install_property(g_iface, pspec.get());
     }
 
-    gjs_hash_table_for_gsize_remove(class_init_properties, gtype);
+    class_init_properties.erase(found);
 }
 
 static void
 gjs_object_class_init(GObjectClass *klass,
                       gpointer      user_data)
 {
-    GPtrArray *properties;
-    GType gtype;
-    guint i;
-
-    gtype = G_OBJECT_CLASS_TYPE (klass);
+    GType gtype = G_OBJECT_CLASS_TYPE(klass);
 
     klass->constructor = gjs_object_constructor;
     klass->set_property = gjs_object_set_gproperty;
     klass->get_property = gjs_object_get_gproperty;
 
-    properties = (GPtrArray*) gjs_hash_table_for_gsize_lookup (class_init_properties, gtype);
-    if (properties != NULL) {
-        for (i = 0; i < properties->len; i++) {
-            GParamSpec *pspec = (GParamSpec*) properties->pdata[i];
-            g_param_spec_set_qdata(pspec, gjs_is_custom_property_quark(), GINT_TO_POINTER(1));
-            g_object_class_install_property (klass, i+1, pspec);
-        }
-        
-        gjs_hash_table_for_gsize_remove (class_init_properties, gtype);
+    auto found = class_init_properties.find(gtype);
+    if (found == class_init_properties.end())
+        return;
+
+    ParamRefArray& properties = found->second;
+    unsigned i = 0;
+    for (ParamRef& pspec : properties) {
+        g_param_spec_set_qdata(pspec.get(), gjs_is_custom_property_quark(),
+                               GINT_TO_POINTER(1));
+        g_object_class_install_property(klass, ++i, pspec.get());
     }
+
+    class_init_properties.erase(found);
 }
 
 static void
@@ -2701,36 +2674,26 @@ save_properties_for_class_init(JSContext       *cx,
                                uint32_t         n_properties,
                                GType            gtype)
 {
-    GPtrArray *properties_native = NULL;
-    guint32 i;
-
-    if (!class_init_properties)
-        class_init_properties = gjs_hash_table_new_for_gsize((GDestroyNotify) g_ptr_array_unref);
-    properties_native = g_ptr_array_new_with_free_func((GDestroyNotify) g_param_spec_unref);
-    for (i = 0; i < n_properties; i++) {
-        JS::RootedValue prop_val(cx);
-
-        if (!JS_GetElement(cx, properties, i, &prop_val)) {
-            g_clear_pointer(&properties_native, g_ptr_array_unref);
+    ParamRefArray properties_native;
+    JS::RootedValue prop_val(cx);
+    JS::RootedObject prop_obj(cx);
+    for (uint32_t i = 0; i < n_properties; i++) {
+        if (!JS_GetElement(cx, properties, i, &prop_val))
             return false;
-        }
+
         if (!prop_val.isObject()) {
-            g_clear_pointer(&properties_native, g_ptr_array_unref);
             gjs_throw(cx, "Invalid parameter, expected object");
             return false;
         }
 
-        JS::RootedObject prop_obj(cx, &prop_val.toObject());
-        if (!gjs_typecheck_param(cx, prop_obj, G_TYPE_NONE, true)) {
-            g_clear_pointer(&properties_native, g_ptr_array_unref);
+        prop_obj = &prop_val.toObject();
+        if (!gjs_typecheck_param(cx, prop_obj, G_TYPE_NONE, true))
             return false;
-        }
-        g_ptr_array_add(properties_native, g_param_spec_ref(gjs_g_param_from_param(cx, prop_obj)));
-    }
-    gjs_hash_table_for_gsize_insert(class_init_properties, (gsize) gtype,
-                                    g_ptr_array_ref(properties_native));
 
-    g_clear_pointer(&properties_native, g_ptr_array_unref);
+        properties_native.emplace_back(g_param_spec_ref(gjs_g_param_from_param(cx, prop_obj)),
+                                       g_param_spec_unref);
+    }
+    class_init_properties[gtype] = std::move(properties_native);
     return true;
 }
 
@@ -2805,6 +2768,30 @@ gjs_register_interface(JSContext *cx,
     return true;
 }
 
+static void
+gjs_object_base_init(void *klass)
+{
+    auto priv = static_cast<ObjectInstance *>(g_type_get_qdata(G_OBJECT_CLASS_TYPE(klass),
+                                              gjs_object_priv_quark()));
+
+    if (priv) {
+        for (GClosure *closure : priv->closures)
+            g_closure_ref(closure);
+    }
+}
+
+static void
+gjs_object_base_finalize(void *klass)
+{
+    auto priv = static_cast<ObjectInstance *>(g_type_get_qdata(G_OBJECT_CLASS_TYPE(klass),
+                                              gjs_object_priv_quark()));
+
+    if (priv) {
+        for (GClosure *closure : priv->closures)
+            g_closure_unref(closure);
+    }
+}
+
 static bool
 gjs_register_type(JSContext *cx,
                   unsigned   argc,
@@ -2818,8 +2805,8 @@ gjs_register_type(JSContext *cx,
     GTypeInfo type_info = {
         0, /* class_size */
 
-	(GBaseInitFunc) NULL,
-	(GBaseFinalizeFunc) NULL,
+        gjs_object_base_init,
+        gjs_object_base_finalize,
 
 	(GClassInitFunc) gjs_object_class_init,
 	(GClassFinalizeFunc) NULL,
@@ -2892,8 +2879,13 @@ gjs_register_type(JSContext *cx,
         gjs_add_interface(instance_type, iface_types[i]);
 
     /* create a custom JSClass */
-    JS::RootedObject module(cx, gjs_lookup_private_namespace(cx)), constructor(cx);
-    gjs_define_object_class(cx, module, NULL, instance_type, &constructor);
+    JS::RootedObject module(cx, gjs_lookup_private_namespace(cx));
+    JS::RootedObject constructor(cx), prototype(cx);
+    gjs_define_object_class(cx, module, nullptr, instance_type, &constructor,
+                            &prototype);
+
+    ObjectInstance *priv = priv_from_js(cx, prototype);
+    g_type_set_qdata(instance_type, gjs_object_priv_quark(), priv);
 
     argv.rval().setObject(*constructor);
 
@@ -3022,3 +3014,16 @@ gjs_lookup_object_constructor(JSContext             *context,
     value_p.setObject(*constructor);
     return true;
 }
+
+bool
+gjs_object_associate_closure(JSContext       *cx,
+                             JS::HandleObject object,
+                             GClosure        *closure)
+{
+    ObjectInstance *priv = priv_from_js(cx, object);
+    if (!priv)
+        return false;
+
+    do_associate_closure(priv, closure);
+    return true;
+}
