@@ -490,10 +490,9 @@ do_import(JSContext       *context,
     JS::RootedValue elem(context);
 
     /* First try importing an internal module like byteArray */
-    if (priv->is_root && gjs_is_registered_native_module(context, obj, name)) {
-        if (!import_native_file(context, obj, name))
-            goto out;
-
+    if (priv->is_root &&
+        gjs_is_registered_native_module(context, obj, name) &&
+        import_native_file(context, obj, name)) {
         gjs_debug(GJS_DEBUG_IMPORTER,
                   "successfully imported module '%s'", name);
         result = true;
@@ -627,7 +626,7 @@ do_import(JSContext       *context,
         /* If no exception occurred, the problem is just that we got to the
          * end of the path. Be sure an exception is set.
          */
-        gjs_throw_custom(context, "Error", "ImportError",
+        gjs_throw_custom(context, JSProto_Error, "ImportError",
                          "No JS module '%s' found in search path", name);
     }
 
@@ -762,10 +761,8 @@ importer_resolve(JSContext        *context,
         return true;
     }
 
-    if (!gjs_get_string_id(context, id, &name)) {
-        *resolved = false;
-        return true;
-    }
+    if (!gjs_get_string_id(context, id, &name))
+        return false;
 
     /* let Object.prototype resolve these */
     if (strcmp(name, "valueOf") == 0 ||
