@@ -180,7 +180,7 @@ void        gjs_throw                        (JSContext       *context,
                                               const char      *format,
                                               ...)  G_GNUC_PRINTF (2, 3);
 void        gjs_throw_custom                 (JSContext       *context,
-                                              const char      *error_class,
+                                              JSProtoKey       error_kind,
                                               const char      *error_name,
                                               const char      *format,
                                               ...)  G_GNUC_PRINTF (4, 5);
@@ -253,11 +253,8 @@ const char* gjs_get_type_name                (JS::Value        value);
 /* Functions intended for more "internal" use */
 
 void gjs_maybe_gc (JSContext *context);
-
-bool gjs_context_get_frame_info(JSContext                             *cx,
-                                mozilla::Maybe<JS::MutableHandleValue> stack,
-                                mozilla::Maybe<JS::MutableHandleValue> fileName,
-                                mozilla::Maybe<JS::MutableHandleValue> lineNumber);
+void gjs_schedule_gc_if_needed(JSContext *cx);
+void gjs_gc_if_needed(JSContext *cx);
 
 bool gjs_eval_with_scope(JSContext             *context,
                          JS::HandleObject       object,
@@ -288,6 +285,7 @@ typedef enum {
   GJS_STRING_STACK,
   GJS_STRING_FILENAME,
   GJS_STRING_LINE_NUMBER,
+  GJS_STRING_COLUMN_NUMBER,
   GJS_STRING_NAME,
   GJS_STRING_X,
   GJS_STRING_Y,
@@ -340,6 +338,22 @@ bool gjs_object_define_property(JSContext       *cx,
                                 JSNative         getter = nullptr,
                                 JSNative         setter = nullptr);
 
+bool gjs_object_define_property(JSContext       *cx,
+                                JS::HandleObject obj,
+                                GjsConstString   property_name,
+                                JS::HandleString value,
+                                unsigned         flags,
+                                JSNative         getter = nullptr,
+                                JSNative         setter = nullptr);
+
+bool gjs_object_define_property(JSContext       *cx,
+                                JS::HandleObject obj,
+                                GjsConstString   property_name,
+                                uint32_t         value,
+                                unsigned         flags,
+                                JSNative         getter = nullptr,
+                                JSNative         setter = nullptr);
+
 JS::HandleId gjs_context_get_const_string(JSContext     *cx,
                                           GjsConstString string);
 
